@@ -1,6 +1,7 @@
-/*  an v0.93 - Anagram generator
+/*  an v0.95 - Anagram generator
     Copyright (C) 1996, Free Software Foundation.
     Copyright (C) 1995,1996  Richard Jones
+    Copyright (C) 2001,2002  Paul Martin <pm@debian.org>
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -27,6 +28,7 @@
 #include <locale.h>
 /*#include <math.h>*/
 #include <getopt.h>
+#include <assert.h>
 
 #ifdef __GLIBC__
 typedef int (*gan_qsort_t) (const void *, const void *);
@@ -63,7 +65,7 @@ typedef int (*gan_qsort_t) (const void *, const void *);
 
 #define max(a,b) ((a) >= (b) ? (a) : (b))
 
-#define VERSION "an v0.94 - Mar 14 2001"
+#define VERSION "an v0.95 - May 6 2002"
 #define TRUE 1
 #define FALSE 0
 
@@ -82,31 +84,30 @@ typedef int (*gan_qsort_t) (const void *, const void *);
 #define BIG_INT (2<<16)		/* Magic */
 
 void lowercase (char *string);
-void remove_punc (char *string);
-int check_letter (char letter, char *phrase_word, int *avail);
-int check_word (char *current_word, char *phrase_word);
+void remove_punc (unsigned char *string);
+int check_letter (unsigned char letter, unsigned char *phrase_word, int *avail);
+int check_word (unsigned char *current_word, unsigned char *phrase_word);
 void check_dict (int phrase_len, bitmask_t *phrase_mask,
 		 int num_phrase_masks, bitmask_t **word_mask,
 		 int level, int entry_point, int old_high_letter_num);
-int remove_letters (char *remove_phrase, char *phrase_word);
-int get_words (char *dict_file, char *phrase_word, FILE * dict);
-void make_word_masks (char **words, bitmask_t **word_mask, int *num_word_masks,
-		      int dict_words);
-int makefreq_table (int *freq_table, char *word);
-//int power (int x, int n);
-//char *tobinary (long int number);
+int remove_letters (unsigned char *remove_phrase, unsigned char *phrase_word);
+int get_words (unsigned char *dict_file, unsigned char *phrase_word,
+	       FILE * dict);
+void make_word_masks (unsigned char **words, bitmask_t **word_mask,
+		      int *num_word_masks, int dict_words);
+int makefreq_table (int *freq_table, unsigned char *word);
 int compare_len (char *string1[], char *string2[]);
-int no_punc_len (char *string);
+int no_punc_len (unsigned char *string);
 void *xmalloc (size_t size);
 void *xcalloc (size_t num_objects, size_t size);
 void *xrealloc (void *location, size_t newsize);
 char *xstrdup (char *s);
 void create_let_hash (bitmask_t **word_mask, int dict_words,
 		      bitmask_t *letter_mask, int *let_hash);
-void calc_letter_freq (char **words, int *letter_freq, int dict_words);
-int compare_char (char *let1, char *let2);
-int compare_word (char *word1[], char *word2[]);
-int freq_strcmp (char *string1, char *string2);
+void calc_letter_freq (unsigned char **words, int *letter_freq, int dict_words);
+int compare_char (unsigned char *let1, unsigned char *let2);
+int compare_word (unsigned char *word1[], unsigned char *word2[]);
+int freq_strcmp (unsigned char *string1, unsigned char *string2);
 void create_letter_index (int *letter_index);
 int compare_let_index (int *let1, int *let2);
 void create_word_len_hash (int **word_len_hash, int *word_len, int *let_hash,
@@ -121,12 +122,12 @@ int length_of_anagram = MAXWORDS;	/* Maximum number of words in anagram */
 int max_num_anagrams = FALSE;	/*Maximum number of anagrams to output */
 int num_anagrams = 0;		/* Number of anagrams found */
 int output_words = FALSE;	/* output words option flag */
-char **words;			/* array of words to make anagrams from */
+unsigned char **words;		/* array of words to make anagrams from */
 int **word_len_hash;		/* length & letter index into words array */
 int maxword_len = 0;		/* Maximum word length in words array */
 
 char length_option = FALSE;
-char *contain_phrase = 0;	/* pointer to contain phrase */
+unsigned char *contain_phrase = 0;	/* pointer to contain phrase */
 
 int *num_word_masks;		/* array of number of bit masks for each word */
 int *let_hash;
@@ -150,8 +151,8 @@ int main (argc, argv)
    int dict_words = 0;		/* number of words in new dictionary */
    int loop1;			/* loop counter */
    int loop2;
-   char *phrase_word = 0;	/* array for users phrase */
-   char *old_phrase_word;	/* PHRASE_WORD before processing */
+   unsigned char *phrase_word = 0;	/* array for users phrase */
+   unsigned char *old_phrase_word;	/* PHRASE_WORD before processing */
    int level = 0;		/* level of current recursion */
    int phrase_len;		/* length of users phrase */
    int *freq_table;		/* array of letter frequencies for current word */
@@ -165,10 +166,10 @@ int main (argc, argv)
    int used_flag = FALSE;	/* used phrase option flag */
    int test_flag = FALSE;	/* test option flag */
    char *dict_file = NULL;	/* pointer to dictionary file name */
-   char *test_phrase = 0;	/* pointer to test phrase */
-   char *old_test_phrase;	/* TEST_PHRASE before processing */
+   unsigned char *test_phrase = 0;	/* pointer to test phrase */
+   unsigned char *old_test_phrase;	/* TEST_PHRASE before processing */
 
-   char *used_phrase = 0;	/* pointer to used phrase */
+   unsigned char *used_phrase = 0;	/* pointer to used phrase */
    int option_index;		/* index into LONG_OPTIONS structure */
    bitmask_t mask;
 
@@ -289,7 +290,7 @@ int main (argc, argv)
 
    letter_mask = xcalloc (alphabet, sizeof(bitmask_t));
    for (loop1 = 0; loop1 < alphabet; loop1++) {
-     letter_mask[loop1] = 1 << (alphabet - loop1 - 1);
+     letter_mask[loop1] = (bitmask_t) 1 << (alphabet - loop1 - 1);
    }
    
    letter_freq = xcalloc (alphabet, sizeof (int));
@@ -474,7 +475,7 @@ int main (argc, argv)
 	 freq_table[loop1]--;
       }
    }
-//   free(freq_table);
+   free(freq_table);
 
 
    /* Allocate memory for and calculate the WORD_LEN_HASH array which
@@ -547,10 +548,15 @@ check_dict (phrase_len, phrase_mask, num_phrase_masks,
    int new;			/* Counter used in NEW_PHRASE_MASK creation */
    int len_check_val = 0;	/* Value used in length short-circuiting */
 #ifdef FORCE_ANSI
-   bitmask_t *new_phrase_mask = xmalloc (num_phrase_masks * sizeof (bitmask_t)); /* New PHRASE_MASK to 
-										   pass to itself */
+   bitmask_t *new_phrase_mask =
+     xmalloc (num_phrase_masks * sizeof (bitmask_t));
+                                /* New PHRASE_MASK to 
+				   pass to itself */
 #endif
 
+   assert(phrase_len>=0);	 
+   /* Compilation with wrong datatype for bitmask_t can trigger this */
+   
    if (length_option)
       len_check_val = length_of_anagram - level - 1;
 
@@ -570,7 +576,8 @@ check_dict (phrase_len, phrase_mask, num_phrase_masks,
 
    {
       int n;			/* Start point in WORDS array */
-      int end;			/* End point in WORDS array */
+      int end=0;		/* End point in WORDS array */
+      int high;			/* Loop counter */
       /* WORDS[] is sorted by letter frequency and length, from longest to 
          shortest, therefore we can jump to the first word of the length of
          the current number of letters left unprocessed in the phrase via the 
@@ -581,10 +588,13 @@ check_dict (phrase_len, phrase_mask, num_phrase_masks,
       if (entry_point > n)
 	 n = entry_point;
 
-      end = let_hash[high_letter_num + 1];	/* Recursive routine only uses words
-						   which have the currently least
-						   frequent letter in them at each
-						   level of recursion */
+      /* Recursive routine only uses words which have the currently least
+         frequent letter in them at each level of recursion */
+
+      for (high = high_letter_num + 1; high < alphabet ; high++) {
+	 end = let_hash[high];
+	 if (end > n) break;
+      }
 
       /* Search through all words containing current least frequent letter
          for anagrams */
@@ -819,8 +829,8 @@ void print (found2, level, upto)
    returns TRUE if it can make it, FALSE otherwise */
 
 int check_word (current_word, phrase_word)
-     char *current_word;
-     char *phrase_word;
+     unsigned char *current_word;
+     unsigned char *phrase_word;
 {
    int loop1;			/* Loop counter */
    int phrase_len = strlen (phrase_word);	/* Length of word checking against */
@@ -832,7 +842,7 @@ int check_word (current_word, phrase_word)
 #endif
 
    int useit;			/* TRUE if we can use current letter */
-   char letter;			/* Current letter being checked */
+   unsigned char letter;	/* Current letter being checked */
 
 
 
@@ -888,8 +898,8 @@ int check_word (current_word, phrase_word)
    allready */
 
 int check_letter (letter, phrase_word, avail)
-     char letter;
-     char *phrase_word;
+     unsigned char letter;
+     unsigned char *phrase_word;
      int *avail;
 {
 
@@ -915,8 +925,8 @@ int check_letter (letter, phrase_word, avail)
 /* Extract letters in REMOVE_PHRASE from PHRASE_WORD */
 
 int remove_letters (remove_phrase, phrase_word)
-     char *remove_phrase;
-     char *phrase_word;
+  unsigned char *remove_phrase;
+  unsigned char *phrase_word;
 {
 
    int loop2;
@@ -969,8 +979,8 @@ int remove_letters (remove_phrase, phrase_word)
    store in WORDS array */
 
 int get_words (dict_file, phrase_word, dict)
-     char *dict_file;
-     char *phrase_word;
+     unsigned char *dict_file;
+     unsigned char *phrase_word;
      FILE *dict;
 {
    int num_words = 0;
@@ -1079,7 +1089,7 @@ int get_words (dict_file, phrase_word, dict)
 /* Create the bit masks for each word in WORDS array */
 
 void make_word_masks (words, word_mask, num_word_masks, dict_words)
-     char **words;
+     unsigned char **words;
      bitmask_t **word_mask;
      int *num_word_masks;
      int dict_words;
@@ -1180,7 +1190,7 @@ void create_let_hash (bitmask_t **word_mask, int dict_words,
 
 int makefreq_table (freq_table, word)
      int *freq_table;
-     char *word;
+     unsigned char *word;
 {
    int max_freq = 1;		/* Maximum number of times any letter appears in WORD */
    int letfreq;			/* Number of times individual letter appears in FREQ_TABLE */
@@ -1218,11 +1228,11 @@ int makefreq_table (freq_table, word)
 
 /* Calculate number of times each letter appears in WORDS array */
 
-void calc_letter_freq (char **words, int *letter_freq, int dict_words)
+void calc_letter_freq (unsigned char **words, int *letter_freq, int dict_words)
 {
 
    int loop1 = 0;
-   char *tmpptr;
+   unsigned char *tmpptr;
    loop1 = dict_words;
 
    while (loop1--) {
@@ -1266,8 +1276,8 @@ int compare_len (string1, string2)
 /* Compare function used by qsort to sort the words in WORDS array */
 
 int compare_word (word1, word2)
-     char *word1[];
-     char *word2[];
+     unsigned char *word1[];
+     unsigned char *word2[];
 {
 
 #ifndef FORCE_ANSI
@@ -1381,8 +1391,8 @@ int compare_word (word1, word2)
    array */
 
 int compare_char (let1, let2)
-     char *let1;
-     char *let2;
+     unsigned char *let1;
+     unsigned char *let2;
 {
 
    /* Check to make sure punctuation appears at the end of words */
@@ -1421,7 +1431,7 @@ int compare_char (let1, let2)
    letter goes before a higher frequency) */
 
 
-int freq_strcmp (char *s1, char *s2)
+int freq_strcmp (unsigned char *s1, unsigned char *s2)
 {
 
    int c1;
@@ -1518,20 +1528,15 @@ void create_word_len_hash (word_len_hash, word_len, let_hash, phrase_len)
      int *let_hash;
      int phrase_len;
 {
-
-
    int old_count = 0;
    int loop1, loop2;
 
-
-
-   for (loop1 = 0; loop1 < alphabet; loop1++) {
+   for (loop1 = 0; loop1 <= alphabet; loop1++) {
       for (loop2 = 0; loop2 <= phrase_len; loop2++) {
 	 word_len_hash[loop2][loop1] = -1;
       }
    }
 
-
    for (loop1 = 0; loop1 < alphabet; loop1++) {
       old_count = let_hash[loop1];
       if (old_count < let_hash[alphabet+1])
@@ -1545,10 +1550,7 @@ void create_word_len_hash (word_len_hash, word_len, let_hash, phrase_len)
       }
    }
 
-
-
    for (loop1 = 0; loop1 < alphabet; loop1++) {
-
       word_len_hash[0][loop1] = let_hash[alphabet+1];
       for (loop2 = 1; loop2 < phrase_len; loop2++) {
 	 if (word_len_hash[loop2][loop1] == -1) {
@@ -1575,9 +1577,9 @@ void lowercase (string)
 /* Remove punctuation from STRING */
 
 void remove_punc (string)
-     char *string;
+  unsigned char *string;
 {
-   char *tmp_string;
+   unsigned char *tmp_string;
 
    /* Loop through letters in TMP_STRING copying non-punctuation letters 
       back into STRING */
@@ -1607,7 +1609,7 @@ void remove_punc (string)
    allready lowercase, returns STRING length */
 
 int no_punc_len (string)
-     char *string;
+  unsigned char *string;
 {
    int length = 0;		/* Length of STRING */
 
